/*:
 * =================================================================
 * PACKAGE:	fstat -- File Status Information
 * C-SRC:	Main Command Body
 * VERSION:	%I% (%E%)
 * =================================================================
 * Copyright (1994): Martin Weitzel, Darmstadt, Germany
 *
 * This file contains most of the main command body for "itime".
 * Only some details about getting the system's correct value for
 * the clock ticks is hidden in the module "clk_tck.c".
*/
#include "compile.h"

const char USAGE[] =
	"%s [ -e[msg] ] [ -c[nnn] ] [ +format ]";

#ifndef NO_BUILTIN_HELP
const char *HELP[] = {
	"-- File Status Information",
	"",
	"This program reads file names from stdin and prints further",
	"information, mainly from the file's i-node.  Output format",
	"can be tailored by a format-string that follows a plus sign.",
	"The options are:",
	"  -e[msg]  message to print for files for which no status",
	"           information can be obtained; if 'msg' is missing,",
	"           such files are silently ignored;",
	"  -c[nnn]  number of bytes to include into the checksum (see",
	"           also %H-place-holder); if 'nnn' is missing or zero,",
	"           the checksum is calculated for the complete file;",
	"Within the format string certain place-holders are recognized:",
	"  %t  file type in octal;",
	"  %T  file type coded as in 'ls -l';",
	"  %V  file type verbose;",
	"  %p  permissions in octal;",
	"  %P  permissions coded as in 'ls -l';",
	"  %d  device number (in decimal);",
	"  %i  inode number (in decimal);",
	"  %n  number of links (in decimal);",
	"  %u  file owner's UID (in decimal);",
	"  %U  file owner by name (according to '/etc/passwd');",
	"  %g  file group's GID (in decimal);",
	"  %G  file group by name (according to '/etc/group');",
	"  %s  file size (number of bytes in decimal);",
	"  %a  access time (seconds since urknall);",
	"  %A  access time formatted 'YYMMDDhhmmss';",
	"  %c  creation time (seconds since urknall);",
	"  %C  creation time formatted 'YYMMDDhhmmss';",
	"  %m  modification time (seconds since urknall);",
	"  %M  modification time formatted 'YYMMDDhhmmss';",
	"  %h  file header (first two bytes in hex);",
	"  %H  checksum (see also option 'c');",
	"  %f  basename part of file name;",
	"  %F  full file name (as read from input);",
	"  %e  reserved for use in error message (see below);",
	"An optional number between '%' and the format specifier that",
	"follows is treated as field width.  Output is right justified",
	"unless the number starts with '-'.  If a value does not fit into",
	"that zone, more room is provided automatically.  To print a",
	"'%'-sign, two such characters must be used in sequence.  Other",
	"characters in the format-string are printed as they stand, except",
	"that some '\\'-escape-sequences are recognized and interpreted as",
	"usual.  If a '\\c' is encountered, further interpretation will",
	"stop and the new-line character that otherwise terminates each",
	"line of output will be suppressed.",
	"Place-holders may also be used in user-defined error messages",
	"(see 'e'-option), but when such a message is printed, only '%f'",
	"and '%F' have meaningful values; in addition, '%e' may be used",
	"to print the current 'errno'-value.",
	"If no format-string is specified as argument, the default format",
	"is: %d@%i %F."
};
#endif

/*
 * configurable constants
*/
#define MAX_FILE_NAME	2048
const char DEFAULT_FMT[] = "%d@%i %F";

/*
 * standard header files
*/
#include <assert.h>
#include <ctype.h>
#include <errno.h>
#include <stdarg.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <time.h>

/*
 * system header files
*/
#include <grp.h>
#include <pwd.h>
#include <sys/stat.h>
#ifdef __cplusplus
#include <osfcn.h>
#endif

/*
 * global variables
*/
const char *PROG = "";		/* name of program */

/*
 * print usage message and terminate the program
*/
void usage(void)
{
	(void) printf("Usage: ");
	(void) printf(USAGE, PROG);
	(void) printf("\n");
	exit(1);
}

/*
 * print help text and terminate the program
*/
#ifndef NO_BUILTIN_HELP
void help(void)
{
	int i;
	(void) printf("\t***** %s %s *****\n\nUSAGE:\n\t", PROG, HELP[0]);
	(void) printf(USAGE, PROG);
	(void) printf("\n\nDESCRIPTION:");
	for (i = 1; i < sizeof HELP / sizeof HELP[0]; ++i)
		(void) printf("\t%s\n", HELP[i]);
}
#endif

/*
 * print warning or error message (terminate the program for error)
*/
/*PRINTFLIKE2*/
void error(
	int xit,		/* exit status (0 for warning only) */
	const char *msg,	/* printf-format */
	...			/* more args according to format */
) {
	va_list ap;

	assert (msg != (const char *)0);

	(void) fprintf(stderr, "%s [%s]: ",
		PROG, (xit == 0) ? "WARNING" : "FATAL");
	va_start(ap, msg);
	(void) vfprintf(stderr, msg, ap);
	va_end(ap);
	(void) fprintf(stderr, "\n");
	if (xit)
		exit(xit);
}

/*
 * decode escape sequence from format string
*/
char getescs(		/* RET: char value denoted by esc sequence */
	const char **cpp	/* ref to char position pointer */
) {
	assert(cpp != (const char **)0);
	assert((*cpp) != (const char *)0);
	assert(*(*cpp) == '\\');

	if (isdigit(*++(*cpp))) {
		/*
		 * extract octal char representation
		*/
		char c = 0;
		int k = 0;
		do {
			c *= 8;
			c += *((*cpp)) - '0';
		} while (isdigit(*++(*cpp)) && ++k < 3);
		return c;
	}
	else {
		/*
		 * special char representation
		 */
		switch (*(*cpp)++) {
		case 'c': return -1;
		case 'f': return '\f';
		case 'n': return '\n';
		case 't': return '\t';
		case 'r': return '\r';
		case '\\': return '\\';
		}
		/*
		 * print non-special escaped char 'as is'
		*/
		return (*cpp)[-1];
	}
}

/*
 * get opt. length and format spec from format string
*/
char getformat(		/* RET: format spec char */
	const char **cpp,	/* ref to char position */
	int *lenp		/* return space for opt length */
) {
	int mflag = 0;		/* minus flag */

	assert(cpp != (const char **)0);
	assert((*cpp) != (const char *)0);
	assert(*(*cpp) == '%');

	if (*++(*cpp) == '-') {
		mflag = 1;
		++(*cpp);
	}
	for ((*lenp) = 0; isdigit(*(*cpp)); ++(*cpp)) {
		(*lenp) *= 10;
		(*lenp) += *(*cpp) - '0';
	}
	if (mflag)
		(*lenp) = -(*lenp);
	return *(*cpp)++;
}

/*
 * generate canonical of file type
*/
char ftype(
	int ft
) {
	switch(ft) {
	case S_IFREG: return '-';
	case S_IFDIR: return 'd';
	case S_IFCHR: return 'c';
	case S_IFBLK: return 'b';
	case S_IFIFO: return 'p';
	}
	error(0, "unknown file type: %03o", ft);
	return '?';
}

/*
 * generate verbose file type
*/
const char *ftverb(
	int ft
) {
	switch(ft) {
	case S_IFREG: return "regular";
	case S_IFDIR: return "directory";
	case S_IFCHR: return "char-special";
	case S_IFBLK: return "block-special";
	case S_IFIFO: return "named-pipe";
	default: return "?";
	}
}

/*
 * generate name of file owner
*/
const char *fowner(
	int id
) {
	const struct passwd *pwp = getpwuid(id);

	if (pwp != (struct passwd *)0)
		return pwp->pw_name;
	error(0, "unknown file owner: %d", id);
	return "?";
}

/*
 * generate name of file group
*/
const char *fgroup(
	int id
) {
	const struct group *grp = getgrgid(id);

	if (grp != (struct group *)0)
		return grp->gr_name;
	error(0, "unknown file group: %d", id);
	return "?";
}

/*
 * generate canonical form of permissions
*/
const char *fperm(
	int pm
) {
	static char rb[10];
	rb[0] = (pm & 00400) ? 'r' : '-';
	rb[1] = (pm & 00200) ? 'w' : '-';
	rb[2] = (pm & 00100)
	      ? (pm & 04000) ? 's' : 'x'
	      : (pm & 04000) ? 'S' : '-';
	rb[3] = (pm & 00040) ? 'r' : '-';
	rb[4] = (pm & 00020) ? 'w' : '-';
	rb[5] = (pm & 00010)
	      ? (pm & 02000) ? 's' : 'x'
	      : (pm & 02000) ? 'S' : '-';
	rb[6] = (pm & 00004) ? 'r' : '-';
	rb[7] = (pm & 00002) ? 'w' : '-';
	rb[8] = (pm & 00001)
	      ? (pm & 01000) ? 't' : 'x'
	      : (pm & 01000) ? 'T' : '-';
	assert(rb[sizeof rb - 1] == '\0');
	return rb;
}

/*
 * determine magic number of file (first two bytes)
*/
unsigned int fmagic(
	const char *fn
) {
	FILE *fp;
	int c;
	unsigned int s = 0;

	assert(fn != (char *)0);

	if ((fp = fopen(fn, "rb")) == (FILE *)0)
		error(255, "fopen(3) failed (errno = %d): %s", errno, fn);
	if ((c = getc(fp)) != EOF) {
		s |= c << 8;
		if ((c = getc(fp)) != EOF)
			s |= c << 0;
	}
	(void) fclose(fp);

	return s;
}

/*
 * calculate checksum over given number of bytes
*/
unsigned long cksum(
	const char *fn,
	long n
) {
	FILE *fp;
	int c;
	long i = 0;
	unsigned long s = 0;

	assert(fn != (char *)0);

	if ((fp = fopen(fn, "rb")) == (FILE *)0)
		error(255, "error %d open()-ing file: %s", errno, fn);
	while ((c = getc(fp)) != EOF) {
		s += c;
		if (++i == n) break;
	}
	(void) fclose(fp);

	return s;
}

/*
 * generate canonical format of time stamp
*/
const char *xtime(
	const time_t *tvp
) {
	struct tm *t = gmtime(tvp);
	static const char TFMT[] = "%02d%02d%02d%02d%02d%02d";
	static char rb[13];
	(void) sprintf(rb, TFMT,
		t->tm_year, t->tm_mon+1, t->tm_mday,
		t->tm_hour, t->tm_min, t->tm_sec);
	assert(rb[sizeof rb - 1] == '\0');
	return rb;
}

/*
 * print formatted value(s)
*/
/*PRINTFLIKE1*/
void printfmt(
	const char *fmt,	/* printf-format */
	...			/* more args according to format */
) {
	va_list ap;

	assert(fmt != (const char *)0);

	va_start(ap, fmt);
	(void) vfprintf(stdout, fmt, ap);
	va_end(ap);
}

/*
 * print output for executed command according to format
*/
void print(
	const char *fmt,		/* format */
	const char *fnam,		/* full file name */
	const char *bnam,		/* last part of file name */
	const struct stat *sp,		/* status of file */
	long ncks_or_errno		/* chksum bytes or errno */
) {
	assert(fmt != (char *)0);
	assert(fnam != (char *)0);
	assert(bnam != (char *)0);
	assert(sp != (struct stat *)0);

	/*
	 * return immediately if format string is empty
	*/
	if (*fmt == '\0') return;

	/*
	 * interpret format string
	*/
	while (*fmt) {
		if (*fmt == '%') {
			int len;
			switch (getformat(&fmt, &len)) {
			case 't':
				printfmt("%*o", len,
					(int)(sp->st_mode & S_IFMT) >> 12);
				break;
			case 'T':
				printfmt("%*c", len,
					ftype(sp->st_mode & S_IFMT));
				break;
			case 'V':
				printfmt("%*s", len,
					ftverb(sp->st_mode & S_IFMT));
				break;
			case 'P':
				printfmt("%*s", len,
					fperm(sp->st_mode & ~S_IFMT));
				break;
			case 'p':
				printfmt("%0*o", len ? len : 4,
					(int)(sp->st_mode & ~S_IFMT));
				break;
			case 'd':
				printfmt("%*d", len, (int)sp->st_dev);
				break;
			case 'i':
				printfmt("%*d", len, (int)sp->st_ino);
				break;
			case 'n':
				printfmt("%*d", len, (int)sp->st_nlink);
				break;
			case 'u':
				printfmt("%*d", len, (int)sp->st_uid);
				break;
			case 'U':
				printfmt("%*s", len, fowner(sp->st_uid));
				break;
			case 'g':
				printfmt("%*d", len, (int)sp->st_gid);
				break;
			case 'G':
				printfmt("%*s", len, fgroup(sp->st_gid));
				break;
			case 's':
				printfmt("%*lu", len,
					(unsigned long)sp->st_size);
				break;
			case 'a':
				printfmt("%*ld", len, (long)sp->st_atime);
				break;
			case 'A':
				printfmt("%*s", len, xtime(&(sp->st_atime)));
				break;
			case 'c':
				printfmt("%*ld", len, (long)sp->st_ctime);
				break;
			case 'C':
				printfmt("%*s", len, xtime(&(sp->st_ctime)));
				break;
			case 'm':
				printfmt("%*ld", len, (long)sp->st_mtime);
				break;
			case 'M':
				printfmt("%*s", len, xtime(&(sp->st_mtime)));
				break;
			case 'h':
				printfmt("%0*X", len, fmagic(fnam));
				break;
			case 'H':
				printfmt("%*lu", len,
					cksum(fnam, ncks_or_errno));
				break;
			case 'f':
				printfmt("%*s", len, bnam);
				break;
			case 'F':
				printfmt("%*s", len, fnam);
				break;
			case 'e':
				printfmt("%*d", len, (int)ncks_or_errno);
				break;
			case '%':
				printfmt("%*c", len, '%');
				break;
			default:
				while (*--fmt != '%')
					; /*empty*/
				error(2, "bad format: %s", fmt);
				/*NOTREACHED*/
			}
		}
		else if (*fmt == '\\') {
			int ch = getescs(&fmt);
			if (ch == -1)
				return;
			printfmt("%c", (char)ch);
		}
		else printfmt("%c", *fmt++);
	}
	printfmt("\n");
}

/*
 * main program function
*/
int main(		/* RET: success or failure */
	int argc,		/* arg count */
	char *argv[]		/* arg vector */
) {
	const char *msg = (const char *)0;
	long nbytes = BUFSIZ;
	const char *fmt = DEFAULT_FMT;
	char file_name[MAX_FILE_NAME];
	char *cp;
	struct stat sbuff;
	int i;

	PROG = argv[0];

#	ifndef NO_BUILTIN_HELP
	/*
	 * check for request to print help
	*/
	if (argc == 2 && !strcmp(argv[1], "-help")) {
		help();
		return 0;
	}
#	endif

	/*
	 * loop over flags, get values
	*/
	for (i = 1; i < argc; ++i) {
		if (*argv[i] == '+') fmt = argv[i]+1;
		else if (*argv[i] == '-') {
			switch (argv[i][1]) {
			case 'e':
				msg = argv[i]+2;
				break;
			case 'c':
				nbytes = atoi(argv[i]+2);
				break;
			default:
				error(3, "invalid flag: %c", argv[i][1]);
			}
		}
		else usage();
	}

	/*
	 * check flag values
	*/
	if (nbytes < 0)
		error(3, "bad value for 'c' option: %ld", nbytes);

	/*
	 * process all lines from stdin
	*/
	while (cp = fgets(file_name, sizeof file_name, stdin)) {
		const char *base_name = cp;
		/*
		 * skip empty lines, trim trailing
		 * newline and find last part of name
		*/
		if (*cp == '\n') {
			printfmt("\n");
			continue;
		}
		while (*cp) {
			if (*cp++ == '/')
				base_name = cp;
		}
		if (*--cp == '\n')
			*cp = '\0';
		else {
			error(0, "command exceeds %d chars: %s...",
				sizeof file_name, file_name);
			while (!feof(stdin) && getc(stdin) != '\n')
				; /*empty*/
			continue;
		}

		/*
		 * stat file and show results
		*/
		if (stat(file_name, &sbuff) != -1)
			print(fmt, file_name, base_name, &sbuff, nbytes);
		else if (msg)
			print(msg, file_name, base_name, &sbuff, (int)errno);
		else
			error(0, "stat(2) failed (errno = %d): %s",
				errno, file_name);
	}

	return 0;
}
